# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef class Error:
    cdef sdbus_h.sd_bus_error _e

    cdef reply_from_exception(self, Message call, Exception exception):
        cdef int errno
        cdef int ret

        #NOTE: sd_bus uses the System.Error prefix for general purpose errors
        cdef bytes err_name = b"System.Error." + \
            exception.__class__.__name__.encode('utf-8')

        cdef bytes err_message = str(exception).encode('utf-8')

        sdbus_h.sd_bus_error_set(&self._e, err_name, err_message)
        ret = sdbus_h.sd_bus_reply_method_error(call.message, &self._e)
        if ret < 0:
            raise SdbusError(f"Failed to send error reply: {errorcode[-ret]}", -ret)



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef int listen_message_handler(sdbus_h.sd_bus_message *m,
        void *userdata, sdbus_h.sd_bus_error *err):
    cdef PyObject *listen_ptr = <PyObject*>userdata
    cdef Listen listen = <Listen>listen_ptr
    cdef Message message = Message()

    message.import_sd_bus_message(m)
    args = message.read(listen.signature)
    if listen.seperate_arguments:
        ensure_future(listen.coroutine(*args), loop=listen.loop)
    else:
        ensure_future(listen.coroutine(args), loop=listen.loop)
    return 1

cdef class Listen:
    cdef bytes match
    cdef object coroutine
    cdef bytes signature
    cdef bool seperate_arguments

    cdef sdbus_h.sd_bus_slot *_slot
    cdef object loop

    def __cinit__(self, Service service, address, path, interface, member,
            coroutine, args=(), signature=b'ANY', seperate_arguments=True):

        match = []
        match.append(f"sender='{address}'")
        match.append("type='signal'")
        match.append(f"interface='{interface}'")
        match.append(f"member='{member}'")
        match.append(f"path='{path}'")
        match += [f"arg{i}='{a}'" for i,a in enumerate(args)]

        self.match = (','.join(match)).encode()
        self.coroutine = coroutine
        self.signature = signature
        self.loop = service.loop
        self.seperate_arguments = seperate_arguments

        ret = sdbus_h.sd_bus_add_match(service.bus, &self._slot, self.match,
            listen_message_handler, <void *>self)
        if ret < 0:
            raise SdbusError(f"Failed to add match: {errorcode[-ret]}", -ret)

    def __dealloc__(self):
        self._slot = sdbus_h.sd_bus_slot_unref(self._slot)



# == Copyright: 2017, CCX Technologies


class VariantWrapper:
    """Creates a D-Bus Variant from any value.
            For sending to D-Bus Interfaces.
    """
    dbus_signature = "v"

    def __init__(self, value):
        self.dbus_value = value
        self.value = value



# == Copyright: 2017, CCX Technologies

from adbus.client.call import call
from adbus.client.getset import get
from adbus.client.getset import set_
from adbus.client.getset import get_all
from adbus.client.listen import Listen
from adbus.client.proxy import Proxy



# Copyright: 2017, CCX Technologies
#cython: language_level=3

from libc cimport stdint

cdef extern from "systemd/sd-bus-protocol.h":

    cdef enum:
        _SD_BUS_TYPE_INVALID         =   0 # NUL
        SD_BUS_TYPE_BYTE             = 121 # 'y'
        SD_BUS_TYPE_BOOLEAN          =  98 # 'b'
        SD_BUS_TYPE_INT16            = 110 # 'n'
        SD_BUS_TYPE_UINT16           = 113 # 'q'
        SD_BUS_TYPE_INT32            = 105 # 'i'
        SD_BUS_TYPE_UINT32           = 117 # 'u'
        SD_BUS_TYPE_INT64            = 120 # 'x'
        SD_BUS_TYPE_UINT64           = 116 # 't'
        SD_BUS_TYPE_DOUBLE           = 100 # 'd'
        SD_BUS_TYPE_STRING           = 115 # 's'
        SD_BUS_TYPE_OBJECT_PATH      = 111 # 'o'
        SD_BUS_TYPE_SIGNATURE        = 103 # 'g'
        SD_BUS_TYPE_UNIX_FD          = 104 # 'h'
        SD_BUS_TYPE_ARRAY            =  97 # 'a'
        SD_BUS_TYPE_VARIANT          = 118 # 'v'
        SD_BUS_TYPE_STRUCT           = 114 # 'r'
        SD_BUS_TYPE_STRUCT_BEGIN     =  40 # '('
        SD_BUS_TYPE_STRUCT_END       =  41 # ')'
        SD_BUS_TYPE_DICT_ENTRY       = 101 # 'e'
        SD_BUS_TYPE_DICT_ENTRY_BEGIN = 123 # '{'
        SD_BUS_TYPE_DICT_ENTRY_END   = 124 # '}'

cdef extern from "systemd/sd-bus.h":

    ctypedef int (*sd_bus_message_handler_t)(sd_bus_message *m,
            void *userdata, sd_bus_error *ret_error)

    ctypedef int (*sd_bus_property_get_t) (sd_bus *bus, const char *path,
            const char *interface, const char *property,
            sd_bus_message *reply, void *userdata, sd_bus_error *ret_error)

    ctypedef int (*sd_bus_property_set_t) (sd_bus *bus, const char *path,
            const char *interface, const char *property,
            sd_bus_message *value, void *userdata, sd_bus_error *ret_error)

ctypedef struct sd_bus_vtable_start:
    size_t element_size

ctypedef struct sd_bus_vtable_method:
    const char *member
    const char *signature
    const char *result
    sd_bus_message_handler_t handler
    size_t offset

ctypedef struct sd_bus_vtable_signal:
    const char *member
    const char *signature

ctypedef struct sd_bus_vtable_property:
    const char *member
    const char *signature
    sd_bus_property_get_t get
    sd_bus_property_set_t set
    size_t offset

ctypedef union sd_bus_vtable_data:
    sd_bus_vtable_start start
    sd_bus_vtable_method method
    sd_bus_vtable_signal signal
    sd_bus_vtable_property property

cdef extern from "systemd/sd-bus-vtable.h":

    cdef enum:
        _SD_BUS_VTABLE_START             = 60 # ascii '<'
        _SD_BUS_VTABLE_END               = 62 # ascii '>'
        _SD_BUS_VTABLE_METHOD            = 77 # ascii 'M'
        _SD_BUS_VTABLE_SIGNAL            = 83 # ascii 'S'
        _SD_BUS_VTABLE_PROPERTY          = 80 # ascii 'P'
        _SD_BUS_VTABLE_WRITABLE_PROPERTY = 87 # ascii 'W'

    cdef enum:
        SD_BUS_VTABLE_DEPRECATED                   = 1ULL << 0
        SD_BUS_VTABLE_HIDDEN                       = 1ULL << 1
        SD_BUS_VTABLE_UNPRIVILEGED                 = 1ULL << 2
        SD_BUS_VTABLE_METHOD_NO_REPLY              = 1ULL << 3
        SD_BUS_VTABLE_PROPERTY_CONST               = 1ULL << 4
        SD_BUS_VTABLE_PROPERTY_EMITS_CHANGE        = 1ULL << 5
        SD_BUS_VTABLE_PROPERTY_EMITS_INVALIDATION  = 1ULL << 6
        SD_BUS_VTABLE_PROPERTY_EXPLICIT            = 1ULL << 7

    ctypedef struct sd_bus_vtable:
        stdint.uint8_t type
        stdint.uint64_t flags
        sd_bus_vtable_data x

cdef extern from "systemd/sd-bus.h":

    cdef enum:
        SD_BUS_NAME_REPLACE_EXISTING  = 1 << 0
        SD_BUS_NAME_ALLOW_REPLACEMENT = 1 << 1
        SD_BUS_NAME_QUEUE             = 1 << 2

    ctypedef struct sd_bus:
        pass

    ctypedef struct sd_bus_message:
        pass

    ctypedef struct sd_bus_slot:
        pass

    ctypedef struct sd_bus_error:
        const char *name
        const char *message
        int _need_free

    ctypedef struct sd_bus_error_map:
        const char* name
        int code

    int sd_bus_open_user(sd_bus **ret)
    int sd_bus_open_system(sd_bus **ret)
    int sd_bus_request_name(sd_bus *bus, const char *name, stdint.uint64_t flags)
    int sd_bus_get_unique_name(sd_bus *bus, const char **unique)
    sd_bus *sd_bus_unref(sd_bus *bus)

    int sd_bus_add_object_vtable(sd_bus *bus, sd_bus_slot **slot,
            const char *path, const char *interface,
            const sd_bus_vtable *vtable, void *userdata)
    sd_bus_slot* sd_bus_slot_unref(sd_bus_slot *slot)
    int sd_bus_add_object_manager(sd_bus *bus, sd_bus_slot **slot,
            const char *path)

    int sd_bus_process(sd_bus *bus, sd_bus_message **r)
    int sd_bus_get_fd(sd_bus *bus)

    int sd_bus_message_new_method_return(sd_bus_message *call,
            sd_bus_message **m)
    int sd_bus_message_new_signal(sd_bus *bus, sd_bus_message **m,
            const char *path, const char *interface, const char *member)
    int sd_bus_message_new_method_call(sd_bus *bus, sd_bus_message **m,
            const char *destination, const char *path, const char *interface,
            const char *member)

    const char *sd_bus_message_get_signature(sd_bus_message *m, int complete)
    int sd_bus_message_read_basic(sd_bus_message *m, char type, void *p)

    const sd_bus_error *sd_bus_message_get_error(sd_bus_message *m)
    int sd_bus_message_get_errno(sd_bus_message *m)

    int sd_bus_message_open_container(sd_bus_message *m, char type,
            const char *contents)
    int sd_bus_message_enter_container(sd_bus_message *m, char type,
            const char *contents)
    int sd_bus_message_at_end(sd_bus_message *m, int complete)
    int sd_bus_message_exit_container(sd_bus_message *m)
    int sd_bus_message_close_container(sd_bus_message *m)

    int sd_bus_call_async(sd_bus *bus, sd_bus_slot **slot, sd_bus_message *m,
            sd_bus_message_handler_t callback, void *userdata,
            stdint.uint64_t usec)

    sd_bus_message* sd_bus_message_ref(sd_bus_message *m)
    sd_bus_message* sd_bus_message_unref(sd_bus_message *m)

    int sd_bus_error_set(sd_bus_error *e, const char *name, const char *message)
    int sd_bus_reply_method_error(sd_bus_message *cal, const sd_bus_error *e)
    int sd_bus_error_copy(sd_bus_error *dest, const sd_bus_error *e)

    int sd_bus_message_append_basic(sd_bus_message *m, char type, const void *p)
    int sd_bus_send(sd_bus *bus, sd_bus_message *m, stdint.uint64_t *cookie)

    int sd_bus_emit_properties_changed_strv(sd_bus *bus, const char *path,
            const char *interface, char **names)

    int sd_bus_add_match(sd_bus *bus, sd_bus_slot **slot, const char *match,
            sd_bus_message_handler_t callback, void *userdata)



# Copyright: 2017, CCX Technologies
"""D-Bus Method"""

import inspect
import threading

from .. import sdbus


class Method:
    """Provides an interface between a D-Bus and a Python Method or Function.

    Though this class can be used on its own it is intended to be used with
    the method decorator applied to methods on a class that is a subclass of
    the main Object type.

    Args:
        callback (function): function that will be called when
            D-Bus method is called, it should use new style type
            annotations to define the types of all arguments,
            and the return value, ie. function(x: int, y: str) -> bool:
            if no type is defined a D-Bus Variant will be used
        name (str): optional, name of the method advertised on the D-Bus
            if not set then will use the callback name
        deprecated (bool): optional, if true object is labelled
            as deprecated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        unprivileged (bool): optional, indicates that this method
            may have to ask the user for authentication before
            executing, which may take a little while
        camel_convert (bool): optional, D-Bus method and property
            names are typically defined in Camel Case, but Python
            methods and arguments are typically defined in Snake
            Case, if this is set the cases will be automatically
            converted between the two
        dont_block (bool): optional, if true the method call will not
            block on the D-Bus, **a value will never be returned**
        threadsafe (bool): optional, if the method isn't threadsafe
            (False) the object will lock, so that only one of the
            objects non-threadsafe methods can run at a time,
            default is True
            **NOTE: this will only work if this is decorating a class
            method, as the lock has to be connected to the method**
        executor (concurrent.futures.Executor): optional, executor
            to schedule callback calls on, None will use the default executor

    Raises:
        BusError: if an error occurs during initialization
    """
    def __init__(
            self,
            callback,
            name=None,
            deprecated=False,
            hidden=False,
            unprivileged=False,
            camel_convert=True,
            dont_block=False,
            threadsafe=True,
            executor=None,
    ):
        if not name:
            name = callback.__name__

        self.callback = callback
        """Function called when this method is called."""

        if camel_convert:
            self.dbus_name = sdbus.snake_to_camel(name)
            """Method name advertised on the D-Bus."""
        else:
            self.dbus_name = name

        self.deprecated = deprecated
        self.hidden = hidden
        self.unprivileged = unprivileged
        self.executor = executor

        self.arg_signature = ''
        sig = inspect.signature(callback)
        for param in sig.parameters.values():
            if param.annotation != inspect.Parameter.empty:
                self.arg_signature += sdbus.dbus_signature(param.annotation)
            else:
                self.arg_signature += sdbus.variant_signature()

        self.dont_block = dont_block
        if self.dont_block:
            self.return_signature = ''
        elif sig.return_annotation != inspect.Parameter.empty:
            self.return_signature = sdbus.dbus_signature(sig.return_annotation)
        else:
            self.return_signature = sdbus.variant_signature()

        self.threadsafe = threadsafe

    def __get__(self, instance, owner):
        if instance is None:
            return self

        d = self

        def mfactory(self, *args, **kwargs):
            return d(self, *args, **kwargs)

        mfactory.__name__ = self.callback.__name__
        return mfactory.__get__(instance, owner)

    def __call__(self, *args, **kwargs):
        if not self.threadsafe:
            obj = self.callback.__self__
            if not hasattr(obj, '_adbus_object_lock'):
                obj._adbus_object_lock = threading.RLock()
            with obj._adbus_object_lock:
                return self.callback(*args, **kwargs)
        else:
            return self.callback(*args, **kwargs)

    def vt(self, instance=None):
        return sdbus.Method(
                self.dbus_name, self.callback, self.arg_signature,
                self.return_signature, self.deprecated, self.hidden,
                self.unprivileged, self.dont_block, instance, self.executor
        )


def method(
        name=None,
        deprecated=False,
        hidden=False,
        unprivileged=False,
        camel_convert=True,
        dont_block=False,
        threadsafe=True,
        executor=None
):
    """D-Bus Method Decorator.

    Note:
        The decorated method should use new style type annotations
        to define the types of all arguments, and the return value,
        ie function(x: int, y: str) -> bool:
        if no type is defined a D-Bus Variant will be used

    Args:
        name (str): optional, if set this name will be used on
            the D-Bus, instead of the decorated function's name
        deprecated (bool): optional, if true object is labelled
            as deprecated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        unprivileged (bool): optional, indicates that this method
            may have to ask the user for authentication before
            executing, which may take a little while
        camel_convert (bool): optional, D-Bus method and property
            names are typically defined in Camel Case, but Python
            methods and arguments are typically defined in Snake
            Case, if this is set the cases will be automatically
            converted between the two
        dont_block (bool): optional, if true the method call will not
            block on the D-Bus, **a value will never be returned**
        threadsafe (bool): optional, if the method isn't threadsafe
            (False) the object will lock, so that only one of the
            objects non-threadsafe methods can run at a time,
            default is True
            **NOTE: this will only work if this is decorating a class
            method, as the lock has to be connected to the method**
        executor (concurrent.futures.Executor): optional, executor
            to schedule callback calls on, None will use the default executor

    Returns:
        Instantiated Method, which can be used to replace a method
        or function.
    """
    def wrapper(function):
        return Method(
                function,
                name,
                deprecated,
                hidden,
                unprivileged,
                camel_convert,
                dont_block=dont_block,
                threadsafe=threadsafe,
                executor=executor
        )

    return wrapper



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef class Object:
    cdef sdbus_h.sd_bus *bus
    cdef sdbus_h.sd_bus_slot *_slot
    cdef sdbus_h.sd_bus_vtable *_vtable
    cdef void **_userdata
    cdef bytes path
    cdef bytes interface
    cdef list vtable
    cdef object loop

    def __cinit__(self, service, path, interface, vtable,
            deprecated=False, hidden=False):
        self.vtable = vtable
        self.path = path.encode()
        self.interface = interface.encode()
        self.bus = (<Service>service).bus
        self.loop = (<Service>service).loop

        self._malloc()
        self._init_vtable(deprecated, hidden)
        self._populate_vtable()
        self._register_vtable()

    def __dealloc__(self):
        self._slot = sdbus_h.sd_bus_slot_unref(self._slot)
        PyMem_Free(self._vtable)
        PyMem_Free(self._userdata)

    def _malloc(self):
        length = len(self.vtable)

        self._vtable = <sdbus_h.sd_bus_vtable *>PyMem_Malloc(
                (length+2)*sizeof(sdbus_h.sd_bus_vtable))
        if not self._vtable:
            raise MemoryError("Failed to allocate vtable")
        memset(self._vtable, 0, (length+2)*sizeof(sdbus_h.sd_bus_vtable))

        self._userdata = <void **>PyMem_Malloc(length*sizeof(void*))
        if not self._userdata:
            raise MemoryError("Failed to allocate userdata")
        memset(self._userdata, 0, length*sizeof(void*))

    def _init_vtable(self, deprecated, hidden):
        length = len(self.vtable)

        self._vtable[0].type = sdbus_h._SD_BUS_VTABLE_START
        self._vtable[0].x.start.element_size = sizeof(self._vtable[0])
        self._vtable[0].flags = 0

        if deprecated:
            self._vtable[0].flags |= sdbus_h.SD_BUS_VTABLE_DEPRECATED

        if hidden:
            self._vtable[0].flags |= sdbus_h.SD_BUS_VTABLE_HIDDEN

        self._vtable[length+1].type = sdbus_h._SD_BUS_VTABLE_END
        self._vtable[length+1].flags = 0

    def _populate_vtable(self):
        for i, v in enumerate(self.vtable):
            if type(v) == Method:
                (<Method>v).set_object(self)
                (<Method>v).populate_vtable(&self._vtable[i+1])
                self._vtable[i+1].x.method.offset = i*sizeof(self._userdata[0])
                self._userdata[i] = (<Method>v).userdata

            elif type(v) == Property:
                (<Property>v).set_object(self)
                (<Property>v).populate_vtable(&self._vtable[i+1])
                self._vtable[i+1].x.method.offset = i*sizeof(self._userdata[0])
                self._userdata[i] = (<Property>v).userdata

            elif type(v) == Signal:
                (<Signal>v).set_object(self)
                (<Signal>v).populate_vtable(&self._vtable[i+1])

    def _register_vtable(self):
        ret = sdbus_h.sd_bus_add_object_vtable(self.bus, &self._slot,
                self.path, self.interface, self._vtable, self._userdata)
        if ret < 0:
            raise SdbusError(
                    "Failed to register vtable at "
					f" {self.path} {self.interface}: {errorcode[-ret]}", -ret)

    async def emit_properties_changed(self, property_names):
        cdef int ret
        cdef char **names = <char**>PyMem_Malloc(
                (len(property_names)+1)*sizeof(char*))
        if not names:
            raise MemoryError("Failed to allocate names")
        memset(names, 0, (len(property_names)+1)*sizeof(char*))

        names[len(property_names)] = NULL
        for i, name in enumerate(property_names):
            names[i] = <bytes>name

        ret = sdbus_h.sd_bus_emit_properties_changed_strv(self.bus, self.path,
            self.interface, names)
        if ret < 0:
            raise SdbusError(f"Failed to emit changed: {errorcode[-ret]}", -ret)

        PyMem_Free(names)



# == Copyright: 2017, CCX Technologies

from .. import sdbus


async def call(
        service,
        address,
        path,
        interface,
        method,
        args=(),
        response_signature=None,
        timeout_ms=30000
):
    """Calls a D-Bus Method in another process.

    This is a co-routine, so must be await-ed from within a asyncio mainloop.

    Args:
        service (adbus.server.Service): service to connect to
        address (str): address (name) of the D-Bus Service to call
        path (str): path of method to call, ie. /com/awesome/Settings1
        interface (str): interface label to call, ie. com.awesome.settings
        method (str): name of the method to call, ie. TestMethod
        args (list or tuple): optional, list of arguments to pass to the method
        response_signature (str): optional, D-Bus Signature of the expected
            response, if None the signature will be created at run-time,
            specifying it here will speed up the message post-processing
        timeout_ms (int): maximum time to wait for a response in milli-seconds

    """

    _response_signature = (
            b"Any"
            if response_signature is None else response_signature.encode()
    )

    call = sdbus.Call(
            service.sdbus, address.encode(), path.encode(), interface.encode(),
            method.encode(), args, _response_signature
    )

    call.send(timeout_ms)
    await call.event.wait()

    if isinstance(call.response, Exception):
        raise call.response
    else:
        return call.response



# Copyright: 2017, CCX Technologies
"""D-Bus Object"""

import asyncio

from .. import sdbus


class Object:
    """Provides an interface between a D-Bus and a Python Object.

    Though this class can be used on its own it is intended to be inherited
    by another object which uses decorators to define methods, properties,
    and signals which are exposed on the D-Bus.

    This object must be added to a service, which provides the D-Bus interface.

    It also provides a Context Manager so that multiple properties can be set,
    or one property can be set multiple times, with a single D-Bus changed
    signal being sent for all property updates once the context is closed.

    Args:
        service (adbus.server.Service): service to connect to
        path (str): path to create on the service
            ie. /com/awesome/Settings
        interface (str): interface label to use for all of this
            objects methods and properties, ie. com.awesome.settings
        vtable (list): optional, list of signals, methods, and
            properties that will be added to the D-Bus Object,
            this list is in addition to the methods, properties, and
            signals which are added to this object using the provided
            decorators and descriptors
        deprecated (bool): optional, if true object is labelled
            as deprecated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        manager (bool): optional, if True add a device manager to this object,
            as defined by the D-Bus Spec from freedesktop.org

    Raises:
        BusError: if an error occurs during initialization

    """

    def __init__(
            self,
            service,
            path,
            interface,
            vtable=(),
            deprecated=False,
            hidden=False,
            manager=False,
    ):
        self._defer_properties = False
        self._deferred_property_signals = {}

        self.service = service
        self.path = path

        vtable = list(vtable)
        vtable += [
                v.vt(self)
                for v in type(self).__dict__.values()
                if hasattr(v, 'vt')
        ]

        self.sdbus = sdbus.Object(
                service.sdbus, path, interface, vtable, deprecated, hidden
        )

        if manager:
            self.manager = sdbus.Manager(service.sdbus, path)

    def emit_property_changed(self, dbus_name):
        if self._defer_properties:
            self._deferred_property_signals[dbus_name.encode()] = True

        elif self.service.is_running():
            asyncio.run_coroutine_threadsafe(
                    self.sdbus.emit_properties_changed([dbus_name.encode()]),
                    loop=self.service.get_loop()
            )

    def defer_property_updates(self, enable):
        if enable:
            self._defer_properties = True

        elif self._defer_properties:
            if not self.service.is_running():
                return

            self._defer_properties = False
            if self._deferred_property_signals:
                asyncio.run_coroutine_threadsafe(
                        self.sdbus.emit_properties_changed(
                                list(self._deferred_property_signals.keys())
                        ),
                        loop=self.service.get_loop()
                )

                self._deferred_property_signals = {}

    def __enter__(self):
        self.defer_property_updates(True)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.defer_property_updates(False)



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef class Signal:
    cdef stdint.uint8_t type
    cdef stdint.uint64_t flags
    cdef sdbus_h.sd_bus_vtable_signal x
    cdef void *userdata
    cdef bytes name
    cdef list signature
    cdef bytes arg_signature
    cdef bool connected
    cdef sdbus_h.sd_bus *bus
    cdef bytes path
    cdef bytes interface

    def __cinit__(self, name, signature=(), deprecated=False, hidden=False):

        self.name = name.encode()
        self.signature = [s.encode() for s in signature]
        self.arg_signature = (''.join(signature)).encode()
        self.type = sdbus_h._SD_BUS_VTABLE_SIGNAL
        self.connected = False

        self.flags = 0
        if deprecated:
            self.flags |= sdbus_h.SD_BUS_VTABLE_DEPRECATED

        if hidden:
            self.flags |= sdbus_h.SD_BUS_VTABLE_HIDDEN

        self.x.member = self.name
        self.x.signature = self.arg_signature

    cdef populate_vtable(self, sdbus_h.sd_bus_vtable *vtable):
        vtable.type = self.type
        vtable.flags = self.flags
        memcpy(&vtable.x, &self.x, sizeof(self.x))

    cdef set_object(self, Object object):
        if self.connected:
            raise SdbusError("Signal already associated")
        self.connected = True
        self.bus = object.bus
        self.path = object.path
        self.interface = object.interface

    def emit(self, *values):
        message = Message()
        message.new_signal(self)
        for i, value in enumerate(values):
            try:
                message.append(self.signature[i], value)
            except IndexError:
                raise SdbusError(
                        f"Signal expects {len(self.signature)} arguments "
                        f"but received {len(values)}.")
        message.send()




# Copyright: 2017, CCX Technologies
"""D-Bus Property"""

from .. import sdbus


class Property:
    """Provides an interface between a D-Bus and a Python Property.

    This class is to be used as a Decorator for arguments (properties) in
    an adbus.server.Object which will be exported via the D-Bus.

    Args:
        default: optional, default setting for the property
        name (str): optional, property name used in the D-Bus, if None the
            property's label will be used
        read_only (bool): if True the property can only be read via D-Bus
            (but can still be set locally)
        constant (bool): if True the property can only be read via D-Bus or
            locally, it also advertises constant in the introspect XML data
        deprecated (bool): optional, if true object is labelled
            as deprecated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        unprivileged (bool): optional, indicates that this method
            may have to ask the user for authentication before
            executing, which may take a little while
        emits_change (bool): optional, when changed emits a signal with
            the new value
        emits_invalidation (bool): optional, when changed emits a signal but
            doesn't provide the new value, can be used for large, complex
            properties, overrides emits_change if it's set
        camel_convert (bool): optional, D-Bus method and property
            names are typically defined in Camel Case, but Python
            methods and arguments are typically defined in Snake
            Case, if this is set the cases will be automatically
            converted between the two
    """

    def __init__(
            self,
            default=None,
            name=None,
            read_only=False,
            constant=False,
            deprecated=False,
            hidden=False,
            unprivileged=False,
            emits_change=True,
            emits_invalidation=False,
            camel_convert=True,
    ):

        self.default = default
        self.dbus_name = name
        self.read_only = read_only or constant
        self.constant = constant
        self.deprecated = deprecated
        self.hidden = hidden
        self.unprivileged = unprivileged
        self.emits_change = emits_change and (not emits_invalidation)
        self.emits_invalidation = emits_invalidation
        self.camel_convert = camel_convert

    def __get__(self, instance, owner):
        try:
            return instance.__dict__[self.py_name]
        except KeyError:
            return self.default

    def __set__(self, instance, value):
        if self.constant:
            raise ValueError("Can't change the value of a constant.")

        if getattr(instance, self.py_name) != value:
            instance.__dict__[self.py_name] = value
            self.emit_changed(instance)

    def emit_changed(self, instance):
        if self.emits_change or self.emits_invalidation:
            instance.emit_property_changed(self.dbus_name)

    def __set_name__(self, owner, name):
        self.py_name = name

        try:
            self.dbus_signature = sdbus.dbus_signature(
                    owner.__annotations__[name]
            )
        except KeyError:
            self.dbus_signature = sdbus.variant_signature()

        if not self.dbus_name:
            self.dbus_name = name

        if self.camel_convert:
            self.dbus_name = sdbus.snake_to_camel(self.dbus_name)

    def vt(self, instance):
        """Interface to sd-bus library"""
        return sdbus.Property(
                self.dbus_name, instance, self.py_name, self.dbus_signature,
                self.read_only, self.deprecated, self.hidden,
                self.unprivileged, self.constant, self.emits_change,
                self.emits_invalidation
        )



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef void _method_message_handler(Method method, Message message):
    cdef Error error
    cdef list args
    cdef object value

    args = message.read(method.arg_signature)

    try:
        if method.instance:
            value = method.callback(<object>(method.instance), *args)
        else:
            value = method.callback(*args)
    except Exception as e:
        method.loop.call_exception_handler({'message': str(e), 'exception': e})
        error = Error()
        try:
            error.reply_from_exception(message, e)
        except SdbusError as e:
            method.loop.call_exception_handler({'message': str(e), 'exception': e})
        return

    message.new_method_return()
    message.append(method.return_signature, value)

    try:
        message.send()
    except SdbusError as e:
        method.loop.call_exception_handler({'message': str(e), 'exception': e})

cdef int method_message_handler(sdbus_h.sd_bus_message *m,
        void *userdata, sdbus_h.sd_bus_error *err):

    cdef PyObject *method_ptr = (((<PyObject**>userdata)[0]))
    cdef Method method = <Method>method_ptr
    cdef Message message = Message()

    message.import_sd_bus_message(m)
    method.loop.run_in_executor(method.executor, _method_message_handler, method, message)
    return 1

cdef class Method:

    cdef stdint.uint8_t type
    cdef stdint.uint64_t flags
    cdef sdbus_h.sd_bus_vtable_method x
    cdef void *userdata
    cdef object callback
    cdef object py_instance
    cdef PyObject *instance
    cdef bytes name
    cdef bytes arg_signature
    cdef bytes return_signature
    cdef bool connected
    cdef object loop
    cdef object executor

    def __cinit__(self, name, callback, arg_signature='', return_signature='',
            deprecated=False, hidden=False, unprivileged=False, no_reply=False, instance=None, executor=None):

        self.name = name.encode()
        self.arg_signature = arg_signature[1:].encode() if instance is not None else arg_signature.encode()
        self.return_signature = return_signature.encode()
        self.callback = callback
        self.py_instance = instance
        self.connected = False
        self.executor = executor

        self.type = sdbus_h._SD_BUS_VTABLE_METHOD

        self.flags = 0
        if no_reply:
            self.flags |= sdbus_h.SD_BUS_VTABLE_METHOD_NO_REPLY

        if deprecated:
            self.flags |= sdbus_h.SD_BUS_VTABLE_DEPRECATED

        if hidden:
            self.flags |= sdbus_h.SD_BUS_VTABLE_HIDDEN

        if unprivileged:
            self.flags |= sdbus_h.SD_BUS_VTABLE_UNPRIVILEGED

        self.x.member = self.name
        self.x.handler = method_message_handler
        self.x.signature = self.arg_signature
        self.x.result = self.return_signature

        self.userdata = <void *>self

    cdef populate_vtable(self, sdbus_h.sd_bus_vtable *vtable):
        vtable.type = self.type
        vtable.flags = self.flags
        memcpy(&vtable.x, &self.x, sizeof(self.x))

    cdef set_object(self, Object object):
        if self.connected:
            raise SdbusError("Method already associated")
        self.connected = True

        # this decrements the counter of py_instance so if it
        # refers to the base object the object won't be held in
        # memory indefinitely (self-referenced)
        self.instance = <PyObject *>self.py_instance
        self.py_instance = None

        self.loop = object.loop
